import { system, world, ItemStack, BlockPermutation} from "@minecraft/server";
class StringList {
    constructor() { this.list = new Set(); }
    add(item) { this.list.add(item); }
    has(item) { return this.list.has(item); }
    remove(item) { this.list.delete(item); }
}

function isDiagonal(direction) {
    const normalized = ((direction % 360) + 360) % 360;
    return (normalized % 90 >= 22.5) && (normalized % 90 < 67.5);
}
function isLeftTrue(direction) {
    return (direction > 0 && direction < 45) || (direction > 90 && direction < 135) || (direction > 180 && direction < 225) || (direction > 270 && direction < 315);
}
function usedReduce(player, isMainHand = true) {
    if (player.getGameMode() === "creative" || !player.isValid()) return;
    const slotName = isMainHand ? "Mainhand" : "Offhand";
    const currentItem = player?.getComponent("minecraft:equippable")?.getEquipment(slotName);
    if (!currentItem) return;
    const newAmount = currentItem.amount > 1 ? currentItem.amount - 1 : 0;
    player?.getComponent("minecraft:equippable")?.setEquipment(slotName, newAmount > 0 ? new ItemStack(currentItem.typeId, newAmount) : undefined)
}
function returnOpposite(d) { return {north:'south',south:'north',east:'west',west:'east'}[d]};
world.beforeEvents.worldInitialize.subscribe(initEvent => {
    initEvent.blockComponentRegistry.registerCustomComponent('deco_head:head', {
        beforeOnPlayerPlace: (e) => {
            let face = e.face.toLowerCase();
            let player = e.player;
            let headPerm;
            if (face === "up" || face === "down") {
                let y = player.getRotation().y;
                let rot = y + 360*(y !== Math.abs(y));
                headPerm = e.permutationToPlace.withState('deco_head:is_diagonal', isDiagonal(rot)).withState('deco_head:is_left', isLeftTrue(rot));
            } else {
                headPerm = e.permutationToPlace.withState('deco_head:is_on_wall', true).withState('minecraft:cardinal_direction', returnOpposite(face))
            }
            e.permutationToPlace = headPerm;
        }
    });
});
const variantHandlers = {
    "axolotl": (entity) => {return `head_deco:axolotl_${["lucy", "cyan", "gold", "brown", "blue"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "bee": (entity) => {return entity.getProperty("minecraft:has_nectar") ? entity.getComponent("minecraft:mark_variant")?.value === 1 ? "head_deco:bee_nectar_head" : "head_deco:bee_angry_nectar_head" : entity.getComponent("minecraft:mark_variant")?.value === 1 ? "head_deco:bee_head" : "head_deco:bee_angry_head"},
    "cat": (entity) => {return `head_deco:cat_${["white", "tuxedo", "red", "siamese", "britishshorthair", "calico", "persian", "ragdoll", "tabby", "black", "jellie"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "pillager": () => {return "head_deco:illager_head"},
    "vindicator": () => {return "head_deco:illager_head"},
    "evocation_illager": () => {return "head_deco:illager_head"},
    "fox": (entity) => {return entity.getComponent("minecraft:variant")?.value === 0 ? "head_deco:fox_head" : "head_deco:whitefox_head"},
    "frog": (entity) => {return `head_deco:frog_${["orange", "green", "white"][+entity.getComponent("minecraft:variant")?.value]}_head`},
    "horse": (entity) => {return `head_deco:horse_${["white", "creamy", "chestnut", "brown", "black", "gray", "darkbrown"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "llama": (entity) => {return `head_deco:llama_${["creamy", "white", "brown", "gray"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "mooshroom": (entity) => {return `head_deco:mooshroom_${["red", "brown"][+entity.getComponent("minecraft:variant")?.value || 0]}_head` },
    "ocelot": () => {return `head_deco:cat_ocelot_head`},
    "piglin_brute": () => {return "minecraft:piglin_head"},
    "panda": (entity) => {return `head_deco:panda${["", "_lazy", "_worried", "_playful", "_brown", "_sneezy", "_aggressive"][+entity.getComponent("minecraft:variant")?.value]}_head`},
    "parrot": (entity) => {return `head_deco:parrot_${["red", "blue", "green","yellow_blue", "grey"][+entity.getComponent("minecraft:variant")?.value  || 0]}_head`},
    "rabbit": (entity) => {return `head_deco:rabbit_${["brown", "white_splotched", "black", "toast", "gold", "salt"][entity.getComponent("minecraft:variant")?.value]}_head`},
    "sheep": (entity) => {return entity.nameTag === "jeb_" ? `head_deco:sheep_rainbow_head` : `head_deco:sheep_${["white", "orange", "magenta", "light_blue", "yellow", "lime", "pink", "gray", "light_gray", "cyan", "purple", "blue", "brown", "green", "red", "black"][+entity.getComponent("minecraft:color")?.value || 0]}_head`},
    "shulker": (entity) => {return `head_deco:shulker_${["black", "red", "green", "brown", "blue", "purple", "cyan", "light_gray", "gray", "pink", "lime", "yellow", "light_blue", "magneta", "orange", "white", "undyed"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "strider": (entity) => {return `head_deco:strider${typeof entity.getComponent("minecraft:is_shaking") !== "undefined" ? "_suffocated" : ""}_head`},
    "trader_llama": (entity) => {return `head_deco:trader_llama_${["creamy", "white", "brown", "gray"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "villager_v2": (entity) => {return `head_deco:villager_${["plains", "desert", "jungle", "savanna", "snow", "swamp", "taiga"][+entity.getComponent("minecraft:mark_variant")?.value  || 0]}_${["unskilled", "farmer", "fisherman", "shepherd", "fletcher", "librarian", "cartographer", "cleric", "armorer", "weaponsmith", "toolsmith", "butcher", "leatherworker", "mason", "nitwit"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "wolf": (entity) => {return `head_deco:wolf${["", "_ashen", "_black", "_chestnut", "_rusty", "_snowy", "_spotted", "_striped", "_woods"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
    "zombie_pigman": () => {return "head_deco:zombified_piglin_head"},
    "zombie_villager_v2": (entity) => {return `head_deco:zombie_villager_${["plains", "desert", "jungle", "savanna", "snow", "swamp", "taiga"][+entity.getComponent("minecraft:mark_variant")?.value  || 0]}_${["unskilled", "farmer", "fisherman", "shepherd", "fletcher", "librarian", "cartographer", "cleric", "armorer", "weaponsmith", "toolsmith", "butcher", "leatherworker", "mason", "nitwit"][+entity.getComponent("minecraft:variant")?.value || 0]}_head`},
};

function resolveVariant(entity) {
    const type = entity.typeId.replace("minecraft:", "");
    const handler = variantHandlers[type];
    return handler ? handler(entity) : `head_deco:${type}_head`;
}
world.afterEvents.playerBreakBlock.subscribe(e => {
    let {player, block, brokenBlockPermutation} = e;
    if (player.getGameMode() === "creative" && brokenBlockPermutation?.type.id === "minecraft:creaking_heart") {
        let aboveBlock = block.above();
        let belowBlock = block.below();
        if (aboveBlock.typeId === "minecraft:pale_oak_log" && belowBlock.typeId === "minecraft:pale_oak_log") {
            if (aboveBlock.permutation.getState("pillar_axis") === "y" && belowBlock.permutation.getState("pillar_axis") === "y") {
                player.dimension.getEntities({location: player.location, maxDistance: 34, type: "minecraft:creaking"}).forEach(creaking => {
                    if (creaking.getProperty("minecraft:creaking_state") === "twitching" && !creaking.hasTag("nohead")) {
                        creaking.addTag("nohead");
                        let chance = dropRate["creaking"];
                        if (Math.random() > chance) return;
                        player.dimension.spawnItem(new ItemStack("head_deco:creaking_head"), creaking.location);
                    }
                })
                
            } 
        }
    };

});
const dropRate = {
    "allay": 0.20,
    "axolotl": 1.00,
    "armadillo": 0.30,
    "bat": 0.20,
    "bee": 0.20,
    "breeze": 0.10,
    "blaze": 0.005,
    "bogged": 0.01,
    "camel": 1.00,
    "cat": 0.33,
    "cave_spider": 0.005,
    "chicken": 0.01,
    "cod": 0.10,
    "cow": 0.01,
    "creaking": 0.10,
    "dolphin": 0.33,
    "donkey": 0.20,
    "drowned": 0.05,
    "elder_guardian": 1.00,
    "enderman": 0.005,
    "endermite": 0.10,
    "evoker": 0.25,
    "fox": 1.00,
    "frog": 1.00,
    "ghast": 0.0625,
    "glow_squid": 0.05,
    "goat": 1.00,
    "guardian": 0.005,
    "hoglin": 0.03,
    "horse": 0.27,
    "husk": 0.06,
    "iron_golem": 0.05,
    "llama": 0.24,
    "magma_cube": 0.005,
    "mooshroom": 1.00,
    "mule": 0.20,
    "ocelot": 0.20,
    "panda": 0.27,
    "parrot": 0.25,
    "phantom": 0.10,
    "pig": 0.01,
    "piglin_brute": 0.01,
    "pillager": 0.025,
    "polar_bear": 0.20,
    "pufferfish": 0.15,
    "rabbit": 0.10,
    "ravager": 0.10,
    "salmon": 0.10,
    "sheep": 0.0175,
    "shulker": 0.05,
    "silverfish": 0.10,
    "slime": 0.005,
    "sniffer": 0.8,
    "snow_golem": 0.05,
    "spider": 0.005,
    "squid": 0.05,
    "stray": 0.05,
    "strider": 0.10,
    "tadpole": 0.10,
    "trader_llama": 0.24,
    "tropicalfish": 0.10,
    "turtle": 0.10,
    "vex": 0.1,
    "villager_v2": 0.5,
    "vindicator": 0.05,
    "wandering_trader": 0.5,
    "warden": 1,
    "witch": 0.05,
    "wolf": 0.10,
    "zoglin": 0.1,
    "zombie_villager_v2": 0.5,
    "zombie_pigman": 0.05
};
let chargeCreepIdList = new StringList();
world.afterEvents.entityDie.subscribe(e => {
    let {damageSource, deadEntity} = e;
    let damagingEntity = damageSource?.damagingEntity;
    if (damagingEntity?.typeId === "minecraft:player") {
        let chance = dropRate[deadEntity.typeId.replace("minecraft:", "")] || 0;
        if (Math.random() > chance) return;
        deadEntity.dimension.spawnItem(new ItemStack(resolveVariant(deadEntity)), deadEntity.location);
    };
    if (damagingEntity?.typeId === "minecraft:creeper" && chargeCreepIdList.has(damagingEntity?.id)) {
        system.runTimeout(() => {chargeCreepIdList.remove(damagingEntity?.id)}, 30);
        let chance = dropRate[deadEntity.typeId.replace("minecraft:", "")] || 0;
        if (chance == 0) return;
        deadEntity.dimension.spawnItem(new ItemStack(resolveVariant(deadEntity)), deadEntity.location);
    }
});
world.afterEvents.dataDrivenEntityTrigger.subscribe((e) => {
    chargeCreepIdList.add(e.entity.id);
}, {
    eventTypes: ["minecraft:become_charged"],
    entityTypes: ["minecraft:creeper"]
});
world.beforeEvents.itemUse.subscribe(e => {
    let {itemStack, source} = e;
    let headSlotItem = source?.getComponent("minecraft:equippable")?.getEquipment("Head");
    if (itemStack.typeId.startsWith('head_deco:') && typeof headSlotItem === 'undefined') {
        system.run(() => {
            source.runCommand(`replaceitem entity @s slot.armor.head 0 ${itemStack.typeId} 1`)
            usedReduce(source);
        });
    };
})